/********************************************************************
 *                                                                  *
 * HAV Library, Copyright 1994-2004 Gromada.com                     *
 *                                                                  *
 * THIS FILE IS PART OF THE "HAV Library" SOURCE CODE.              *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'LICENSE.TXT'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.   *
 *                                                                  *
 ********************************************************************
 *                                                                  *
 * hav.c - file format routines                                     *
 *                                                                  *
 ********************************************************************/


#include  <stdlib.h>
#include  <string.h>

#include  "hav.h"


Hav  *hav_handle_alloc (void)
{
  Hav  *havP;

  havP = (Hav *) malloc (sizeof(Hav));
  memset (havP, 0, sizeof(Hav));
  return (havP);
}


void  hav_handle_free (
  Hav *havP)
{
  if (havP) {
    if (havP->palP)
      free (havP->palP);
    free (havP);
    havP = NULL;
  }
}


static void  hav_read_FPT_ (
  Hav        *havP,
  HavStream  *s)
{
  int  n;

  for (n = 0; n < HAV_FPT_ENTRIES; n++)
    f_read (havP->f, &s->FPT.filePos[n], sizeof(s->FPT.filePos[0]));
  f_read (havP->f, &s->FPT.nextFPT, sizeof(s->FPT.nextFPT));
}


HavError  hav_open (
  Hav         *havP, 
  const char  *fileName)
{
  char    havID[4];
  Uint32  hdrSize;

  memset (havP, 0, sizeof(Hav));
  if (!(havP->f = f_open (fileName, F_READ_ONLY)))
    return (E_HAV_FOPEN);
  if (f_read (havP->f, &havID, sizeof(havID)) != sizeof (havID))
    return (E_HAV_FREAD);
  if (strcmp (havID, HAV_ID) != 0)
    return (E_HAV_FORMAT);
  if (f_read (havP->f, &hdrSize, sizeof(hdrSize)) != sizeof (hdrSize))
    return (E_HAV_FREAD);
  f_seek (havP->f, 0, SEEK_SET);
  if (f_read (havP->f, &havP->hdr, hdrSize) != (int)hdrSize)
    return (E_HAV_FREAD);
  if (havP->hdr.minVer > HAV_VER)
    return (E_HAV_MINVER);
  if (havP->hdr.vLen > 0) {
    havP->vFrameSize = havP->hdr.vResX * havP->hdr.vResY * ((havP->hdr.vBpp + 7) / 8);
    havP->vFrameRate = (double)havP->hdr.vFrameRateInt + ((double)havP->hdr.vFrameRateFrac / HAV_FRAME_RATE_FRAC_DIVISOR);
    f_seek (havP->f, havP->hdr.vFirstFPT, SEEK_SET);
    hav_read_FPT_ (havP, &havP->v);
  }
  if (havP->hdr.aLen) {
    havP->aSampleSize = havP->hdr.aChannels * (havP->hdr.aBitsChn / 8);
    havP->aBlocks = (havP->hdr.aLen + havP->hdr.aSamplesPerBlock - 1) / havP->hdr.aSamplesPerBlock;
    f_seek (havP->f, havP->hdr.aFirstFPT, SEEK_SET);
    hav_read_FPT_ (havP, &havP->a);
  }
  return (E_HAV_OK);
}


void  hav_video_get_format (
  Hav     *havP, 
  Uint32  *numFrames,
  Uint32  *codecId, 
  Uint32  *resX,
  Uint32  *resY,
  Uint32  *bpp,
  double  *frameRate,
  Int32   *keyFrameEvery)
{
  if (!havP)
    return;
  *numFrames = havP->hdr.vLen;
  *codecId = havP->hdr.vCodecId;
  *resX = havP->hdr.vResX;
  *resY = havP->hdr.vResY;
  *bpp = havP->hdr.vBpp;
  *frameRate = havP->vFrameRate;
  *keyFrameEvery = havP->hdr.vKeyFrameEvery;
}


void  hav_audio_get_format (
  Hav     *havP, 
  Uint32  *numSamples,
  Uint32  *codecId,
  Uint32  *sampleRate,
  Uint32  *channels, 
  Uint32  *bitsChn)
{
  if (!havP) return;
  *numSamples = havP->hdr.aLen;
  *codecId = havP->hdr.aCodecId;
  *sampleRate = havP->hdr.aSampleRate;
  *channels = havP->hdr.aChannels;
  *bitsChn = havP->hdr.aBitsChn;
}


void  hav_video_seek (
  Hav     *havP,
  Uint32  frameNum)
{
  if (frameNum >= havP->hdr.vLen)
    frameNum = havP->hdr.vLen - 1;
  if (frameNum != havP->v.curBlock) {
    havP->vSkipFrames = frameNum % havP->hdr.vKeyFrameEvery;
    havP->v.curBlock = frameNum - havP->vSkipFrames;
    havP->vSkipFrames--;
  }
  else
    havP->vSkipFrames = 0;
}


static void  hav_seek_fix_ (
  Hav        *havP,
  HavStream  *s,
  Int64      firstFPT)
{
  Uint  newTable;

  newTable = s->curBlock / HAV_FPT_ENTRIES;
  if (newTable < s->curFPT) {
    f_seek (havP->f, firstFPT, SEEK_SET);
    hav_read_FPT_ (havP, s);
    s->curFPT = 0;
  }
  while (s->curFPT < newTable) {
    f_seek (havP->f, s->FPT.nextFPT, SEEK_SET);
    hav_read_FPT_ (havP, s);
    s->curFPT++;
  }
}


static HavError  hav_read_header_ (
  Hav        *havP,
  HavHeader  *hdrP)
{
  Uint32  outputSize;

  outputSize = f_read (havP->f, hdrP->packetId, 4);
  outputSize += f_read (havP->f, &hdrP->hdrSize, sizeof(hdrP->hdrSize));
  outputSize += f_read (havP->f, &hdrP->hdrVer, sizeof(hdrP->hdrVer));
  outputSize += f_read (havP->f, &hdrP->packetSize, sizeof(hdrP->packetSize));
  if (outputSize != HAV_HEADER_SIZE)
    return (E_HAV_FREAD);
  return (E_HAV_OK);
}


static HavError  hav_read_video_header_ (
  Hav      *havP,
  Pointer  headerP)
{
  HavVideoHeader  *hdrP = (HavVideoHeader *) headerP;
  HavError        err; 
  Uint32          outputSize;

  err = hav_read_header_ (havP, &hdrP->hdr);
  if (err != E_HAV_OK)
    return (err);
  outputSize = f_read (havP->f, &hdrP->dataSize, sizeof(hdrP->dataSize));
  outputSize += f_read (havP->f, &hdrP->codecId, sizeof(hdrP->codecId));
  outputSize += f_read (havP->f, &hdrP->palPresent, sizeof(hdrP->palPresent));
  outputSize += f_read (havP->f, &hdrP->palOffset, sizeof(hdrP->palOffset));
  if (outputSize != HAV_VIDEO_HEADER_SIZE - HAV_HEADER_SIZE)
    return (E_HAV_FWRITE);
  return (E_HAV_OK);
}


static HavError  hav_read_palette_packet_ (
  Hav      *havP,
  Pointer  headerP,
  Pointer  dataP)
{
  HavPaletteHeader  *hdrP = (HavPaletteHeader *) headerP;
  HavError          err;
  Uint32            outputSize;

  err = hav_read_header_ (havP, &hdrP->hdr);
  if (err != E_HAV_OK)
    return (err);
  outputSize = f_read (havP->f, &hdrP->firstEntry, sizeof(hdrP->firstEntry));
  outputSize += f_read (havP->f, &hdrP->numOfEntries, sizeof(hdrP->numOfEntries));
  outputSize += f_read (havP->f, dataP, hdrP->numOfEntries * HAV_PALETTE_ENTRY_SIZE);
  if (outputSize != HAV_PALETTE_HEADER_SIZE - HAV_HEADER_SIZE + hdrP->numOfEntries * HAV_PALETTE_ENTRY_SIZE)
    return (E_HAV_FWRITE);
  return (E_HAV_OK);
}


static HavError  hav_video_read_ (
  Hav      *havP)
{
  HavVideoHeader    imgHdr;
  HavPaletteHeader  palHdr;
  Uchar             *cmpsDataP;
  HavPalette        *palDataP;

  if (!havP)
    return (E_HAV_NODATA);
  if (havP->hdr.vCodecId == HAV_NONE)
    return (E_HAV_NODATA);
  if (havP->vPrevImgDataP == NULL)
    havP->vPrevImgDataP = (Pointer) malloc (havP->vFrameSize);
  hav_seek_fix_ (havP, &havP->v, havP->hdr.vFirstFPT);
  f_seek (havP->f, havP->v.FPT.filePos[havP->v.curBlock % HAV_FPT_ENTRIES], SEEK_SET);
  hav_read_video_header_ (havP, &imgHdr);
  f_seek (havP->f, havP->v.FPT.filePos[havP->v.curBlock % HAV_FPT_ENTRIES] + imgHdr.hdr.hdrSize, SEEK_SET);
  if (imgHdr.codecId == HAV_VIDEO_CODEC_RAW) {
    if (f_read (havP->f, havP->vPrevImgDataP, imgHdr.dataSize) != (int)imgHdr.dataSize)
      return (E_HAV_FREAD);
  }
  else {
    if (!(cmpsDataP = (Uchar *) malloc (imgHdr.dataSize)))
      return (E_HAV_MALLOC);
    if (f_read (havP->f, cmpsDataP, imgHdr.dataSize) != (int)imgHdr.dataSize)
      return (E_HAV_FREAD);
    hav_decode ((Uchar *)havP->vPrevImgDataP, cmpsDataP, havP->hdr.vResX, havP->hdr.vResY, havP->hdr.vBpp, imgHdr.codecId);
    free (cmpsDataP);
  }
  if (imgHdr.palPresent) {
    if (havP->palP == NULL)
      if (!(havP->palP = (HavPalette *) malloc (HAV_PALETTE_SIZE)))
        return (E_HAV_MALLOC);
    palDataP = (HavPalette *) malloc (HAV_PALETTE_SIZE);
    hav_read_palette_packet_ (havP, &palHdr, palDataP);
    memmove (&havP->palP[palHdr.firstEntry], palDataP, palHdr.numOfEntries * HAV_PALETTE_ENTRY_SIZE);
    havP->palRefreshed = TRUE;
  }
  else
    havP->palRefreshed = FALSE;
  havP->v.curBlock++;
  return (E_HAV_OK);
}


HavError  hav_video_read (
  Hav      *havP,
  Pointer  imgDataP)
{
  HavError  result;

  while (havP->vSkipFrames > 0) {
    result = hav_video_read_ (havP);
    if (result != E_HAV_OK)
      return (result);
    havP->vSkipFrames--;
  }
  result = hav_video_read_ (havP);
  memmove (imgDataP, havP->vPrevImgDataP, havP->vFrameSize);
  return (result);
}


Bool32  hav_video_get_palette (
  Hav      *havP, 
  Pointer  palDataP,
  Bool     *palRefreshed)
{
  *palRefreshed = havP->palRefreshed;
  if (havP->palP) {
    memmove (palDataP, havP->palP, HAV_PALETTE_SIZE);
    return (TRUE);
  }
  return (FALSE);
}


void  hav_audio_seek (
  Hav     *havP,
  Uint32  sample)
{
  if (sample >= havP->hdr.aLen) {
    havP->a.curBlock = havP->aBlocks - 1;
    havP->aSkipBytes = (havP->hdr.aSamplesPerBlock - 1) * havP->aSampleSize;
  }
  else {
    havP->a.curBlock = sample / havP->hdr.aSamplesPerBlock;
    havP->aSkipBytes = (sample % havP->hdr.aSamplesPerBlock) * havP->aSampleSize;
    havP->aBufPos = 0;
  }
}


static HavError  hav_read_audio_packet_ (
  Hav      *havP,
  Pointer  headerP,
  Pointer  dataP)
{
  HavAudioHeader  *hdrP = (HavAudioHeader *) headerP;
  HavError          err;
  Uint32            outputSize;

  err = hav_read_header_ (havP, &hdrP->hdr);
  if (err != E_HAV_OK)
    return (err);
  outputSize = f_read (havP->f, &hdrP->dataSize, sizeof(hdrP->dataSize));
  outputSize += f_read (havP->f, &hdrP->codecId, sizeof(hdrP->codecId));
  outputSize += f_read (havP->f, &hdrP->numSamples, sizeof(hdrP->numSamples));
  outputSize += f_read (havP->f, dataP, hdrP->dataSize);
  if (outputSize != HAV_AUDIO_HEADER_SIZE - HAV_HEADER_SIZE + hdrP->dataSize)
    return (E_HAV_FWRITE);
  return (E_HAV_OK);
}


HavError  hav_audio_read (
  Hav      *havP, 
  Pointer  sndDataP,
  Uint32   numSamples,
  Uint32   *samplesOut)
{
  HavAudioHeader  sndHdr;
  int             copyBytes;
  int             bytesOut = 0;
  int             bytesToGet = numSamples * havP->aSampleSize;

  *samplesOut = 0;
  if (!havP)
    return (E_HAV_NODATA);
  if (havP->hdr.aCodecId == HAV_NONE)
    return (E_HAV_NODATA);
  if (havP->aBuf == NULL)
    havP->aBuf = (Pointer) malloc (havP->hdr.aSamplesPerBlock * havP->aSampleSize);
  if (havP->aBufPos > 0) {
    copyBytes = havP->aBufUsed - havP->aBufPos;
    if (copyBytes >= bytesToGet) {
      memmove ((Uchar *)sndDataP, (Uchar *)havP->aBuf + havP->aBufPos, bytesToGet);
      if (copyBytes == bytesToGet)
        havP->aBufPos = 0;
      else
        havP->aBufPos += bytesToGet;
      *samplesOut = numSamples;
      return (E_HAV_OK);
    }
    memmove ((Uchar *)sndDataP, (Uchar *)havP->aBuf + havP->aBufPos, copyBytes);
    bytesOut = copyBytes;
    havP->aBufPos = 0;
  }
  while (bytesOut < bytesToGet) {
    if (havP->a.curBlock >= havP->aBlocks)
      break;
    hav_seek_fix_ (havP, &havP->a, havP->hdr.aFirstFPT);
    f_seek (havP->f, havP->a.FPT.filePos[havP->a.curBlock % HAV_FPT_ENTRIES], SEEK_SET);
    hav_read_audio_packet_ (havP, &sndHdr, havP->aBuf);
    havP->a.curBlock++;
    havP->aBufUsed = sndHdr.dataSize;
    if (havP->aSkipBytes >= havP->aBufUsed) {
      havP->aSkipBytes -= havP->aBufUsed;
      continue;
    }
    copyBytes = havP->aBufUsed - havP->aSkipBytes;
    if ((bytesOut + copyBytes) > bytesToGet) {
      copyBytes = bytesToGet - bytesOut;
      havP->aBufPos = copyBytes + havP->aSkipBytes;
    }
    memmove ((Uchar *)sndDataP + bytesOut, (Uchar *)havP->aBuf + havP->aSkipBytes, copyBytes);
    bytesOut += copyBytes;
    havP->aSkipBytes = 0;
  } 
  *samplesOut = bytesOut / havP->aSampleSize;
  return (E_HAV_OK);
}


HavError  hav_create (
  Hav         *havP, 
  const char  *fileName)
{
  memset (havP, 0, sizeof(Hav));
  strcpy (havP->hdr.hdrID, HAV_ID);
  havP->hdr.hdrSize = sizeof(havP->hdr);
  havP->hdr.fmtVer = HAV_VER;
  havP->hdr.minVer = HAV_VER;
  if (!(havP->f = f_open (fileName, F_CREATE)))
    return (E_HAV_FCREATE);
  if (f_write (havP->f, &havP->hdr, sizeof(havP->hdr)) != sizeof (havP->hdr))
    return (E_HAV_FWRITE);
  return (E_HAV_OK);
}


void  hav_video_set_format (
  Hav     *havP, 
  Uint32  codecId, 
  Uint32  resX, 
  Uint32  resY, 
  Uint32  bpp,
  double  frameRate, 
  Int32   keyFrameEvery)
{
  if (!havP) 
    return;
  havP->hdr.vCodecId = codecId;
  havP->hdr.vResX = resX;
  havP->hdr.vResY = resY;
  havP->hdr.vBpp = bpp;
  havP->hdr.vFrameRateInt = (Uint32) frameRate;
  havP->hdr.vFrameRateFrac = (Uint32) ((frameRate - havP->hdr.vFrameRateInt) * HAV_FRAME_RATE_FRAC_DIVISOR);
  havP->hdr.vKeyFrameEvery = (keyFrameEvery == HAV_DEFAULT) ? havP->hdr.vFrameRateInt : (Uint32)keyFrameEvery;
  havP->vFrameSize = resX * resY * ((bpp + 7) / 8);
}


void  hav_audio_set_format (
  Hav     *havP, 
  Uint32  codecId, 
  Uint32  sampleRate,
  Uint32  channels, 
  Uint32  bitsChn,
  Int32   samplesPerBlock)
{
  if (!havP) return;
  havP->hdr.aCodecId = codecId;
  havP->hdr.aSampleRate = sampleRate;
  havP->hdr.aChannels = channels;
  havP->hdr.aBitsChn = bitsChn;
  havP->hdr.aSamplesPerBlock = (samplesPerBlock == HAV_DEFAULT) ? (int)sampleRate : samplesPerBlock;
  havP->aSampleSize = channels * (bitsChn / 8);
}


static void  hav_write_FPT_ (
  Hav        *havP,
  HavStream  *s)
{
  int  n;
  
  f_seek (havP->f, s->FPTFilePos, SEEK_SET);
  for (n = 0; n < HAV_FPT_ENTRIES; n++)
    f_write (havP->f, &s->FPT.filePos[n], sizeof(s->FPT.filePos[0]));
  f_write (havP->f, &s->FPT.nextFPT, sizeof(s->FPT.nextFPT));
}


static void  hav_write_adjust_FPT_ (
  Hav        *havP,
  HavStream  *s,
  Int64      *firstFPT)
{
  if (!s->streamCreated) {
    *firstFPT = f_tell (havP->f);
    s->FPTFilePos = *firstFPT;
    f_seek (havP->f, s->FPTFilePos + HAV_FPT_SIZE, SEEK_SET);
  }
  if (s->FPTPos == HAV_FPT_ENTRIES) {
    s->FPT.nextFPT = f_tell (havP->f);
    hav_write_FPT_ (havP, s);
    s->FPTFilePos = s->FPT.nextFPT;
    f_seek (havP->f, s->FPTFilePos + HAV_FPT_SIZE, SEEK_SET);
    memset (&s->FPT, 0, sizeof(s->FPT));
    s->FPTPos = 0;
  }
  s->FPT.filePos[s->FPTPos] = f_tell (havP->f);
  s->FPTPos++;
}


static void  hav_header_init_ (
  HavHeader   *hdrP,
  Uint        hdrSize,
  const char  *packetId)
{
  strcpy (hdrP->packetId, packetId);
  hdrP->hdrSize = hdrSize;
  hdrP->hdrVer = HAV_VER;
  hdrP->packetSize = hdrSize;
}


static void  hav_video_header_init_ (
  HavVideoHeader  *hdrP,
  Uint32          codecId)
{
  hav_header_init_ ((HavHeader *) hdrP, HAV_VIDEO_HEADER_SIZE, HAV_VIDEO_ID);
  hdrP->dataSize = 0;
  hdrP->codecId = codecId;
  hdrP->palPresent = FALSE;
  hdrP->palOffset = 0;
}


static void  hav_palette_header_init_ (
  HavPaletteHeader  *hdrP,
  Uint32            firstEntry,
  Uint32            numOfEntries)
{
  hav_header_init_ ((HavHeader *) hdrP, HAV_PALETTE_HEADER_SIZE, HAV_PALETTE_ID);
  hdrP->firstEntry = firstEntry;
  hdrP->numOfEntries = numOfEntries;
}


static HavError  hav_write_header_ (
  Hav        *havP,
  HavHeader  *hdrP,
  Uint32     *outputSize)
{
  *outputSize = f_write (havP->f, hdrP->packetId, 4);
  *outputSize += f_write (havP->f, &hdrP->hdrSize, sizeof(hdrP->hdrSize));
  *outputSize += f_write (havP->f, &hdrP->hdrVer, sizeof(hdrP->hdrVer));
  *outputSize += f_write (havP->f, &hdrP->packetSize, sizeof(hdrP->packetSize));
  if (*outputSize != HAV_HEADER_SIZE)
    return (E_HAV_FWRITE);
  return (E_HAV_OK);
}


static HavError  hav_video_write_packet_ (
  Hav      *havP,
  Pointer  headerP,
  Pointer  dataP,
  Uint32   *outputSize)
{
  HavVideoHeader  *hdrP = (HavVideoHeader *) headerP;

  hav_write_header_ (havP, &hdrP->hdr, outputSize);
  *outputSize += f_write (havP->f, &hdrP->dataSize, sizeof(hdrP->dataSize));
  *outputSize += f_write (havP->f, &hdrP->codecId, sizeof(hdrP->codecId));
  *outputSize += f_write (havP->f, &hdrP->palPresent, sizeof(hdrP->palPresent));
  *outputSize += f_write (havP->f, &hdrP->palOffset, sizeof(hdrP->palOffset));
  *outputSize += f_write (havP->f, dataP, hdrP->dataSize);
  if (*outputSize != HAV_VIDEO_HEADER_SIZE + hdrP->dataSize)
    return (E_HAV_FWRITE);
  return (E_HAV_OK);
}


static HavError  hav_palette_write_packet_ (
  Hav      *havP,
  Pointer  headerP,
  Pointer  dataP,
  Uint32   *outputSize)
{
  HavPaletteHeader  *hdrP = (HavPaletteHeader *) headerP;
  Uint32            palSize;

  hav_write_header_ (havP, &hdrP->hdr, outputSize);
  *outputSize += f_write (havP->f, &hdrP->firstEntry, sizeof(hdrP->firstEntry));
  *outputSize += f_write (havP->f, &hdrP->numOfEntries, sizeof(hdrP->numOfEntries));
  palSize = hdrP->numOfEntries * HAV_PALETTE_ENTRY_SIZE;
  *outputSize += f_write (havP->f, dataP, palSize);
  if (*outputSize != HAV_PALETTE_HEADER_SIZE + palSize)
    return (E_HAV_FWRITE);
  return (E_HAV_OK);
}


HavError  hav_video_write_ex (
  Hav      *havP, 
  Pointer  imgDataP, 
  Pointer  palDataP,
  Uint32   *outputSize)
{
  Uint32            cmpsSize, cmpsSizePredict, palValid;
  HavVideoHeader    imgHdr;
  HavPaletteHeader  palHdr;
  Uchar             *dataP;
  Uint32            oSize, trueCodecId;
  HavError          err;

  if ((!havP) || (!imgDataP))
    return (E_HAV_NODATA);
  if (outputSize != NULL)
    *outputSize = 0;
  hav_write_adjust_FPT_ (havP, &havP->v, &havP->hdr.vFirstFPT);
  if (!(dataP = (Uchar *) malloc (havP->vFrameSize)))
    return (E_HAV_MALLOC);
  cmpsSize = hav_encode (dataP, imgDataP, (havP->hdr.vLen % havP->hdr.vKeyFrameEvery != 0) ? havP->vPrevImgDataP : NULL,
    havP->hdr.vResX, havP->hdr.vResY, havP->hdr.vBpp, havP->hdr.vCodecId, &trueCodecId);
  if ((havP->hdr.vCodecId == HAV_VIDEO_CODEC_BEST) && (havP->hdr.vLen == 0)) {
    cmpsSizePredict = hav_encode (dataP, imgDataP, NULL, havP->hdr.vResX, havP->hdr.vResY, havP->hdr.vBpp,
      HAV_VIDEO_CODEC_BEST_PREDICT, &trueCodecId);
    if (cmpsSizePredict < cmpsSize) {
      havP->hdr.vCodecId = HAV_VIDEO_CODEC_BEST_PREDICT;
      cmpsSize = cmpsSizePredict;
    }
    else
      cmpsSize = hav_encode (dataP, imgDataP, NULL, havP->hdr.vResX, havP->hdr.vResY, havP->hdr.vBpp,
        HAV_VIDEO_CODEC_BEST, &trueCodecId);
  }
  hav_video_header_init_ (&imgHdr, trueCodecId);
  imgHdr.dataSize = cmpsSize;
  imgHdr.hdr.packetSize += cmpsSize;
  palValid = (havP->hdr.vBpp == HAV_VIDEO_BITS_8) && (palDataP);
  if (palValid) {
    if (havP->palP)
      palValid = memcmp (havP->palP, palDataP, HAV_PALETTE_SIZE) != 0;
    if (palValid) {
      imgHdr.palPresent = TRUE;
      havP->palOffset = f_tell (havP->f) + imgHdr.hdr.packetSize;
      if (havP->palP == NULL) {
        havP->palP = (HavPalette *) malloc (HAV_PALETTE_SIZE);
        memmove (havP->palP, palDataP, HAV_PALETTE_SIZE);
      }
    }
    imgHdr.palOffset = havP->palOffset;
  }
  err = hav_video_write_packet_ (havP, &imgHdr, dataP, &oSize);
  free (dataP);
  if (err != E_HAV_OK)
    return (err);
  if (outputSize != NULL)
    *outputSize += oSize;
  if (palValid) {
    hav_palette_header_init_ (&palHdr, 0, 256);
    imgHdr.hdr.packetSize += HAV_PALETTE_SIZE;
    err = hav_palette_write_packet_ (havP, &palHdr, palDataP, &oSize);
    if (err != E_HAV_OK)
      return (err);
    if (outputSize != NULL)
      *outputSize += oSize;
  }
  havP->hdr.vLen++;
  if (havP->vPrevImgDataP == NULL)
    havP->vPrevImgDataP = (Pointer) malloc (havP->vFrameSize);
  memmove (havP->vPrevImgDataP, imgDataP, havP->vFrameSize);
  havP->v.streamCreated = TRUE;
  return (E_HAV_OK);
}


HavError  hav_video_write (
  Hav      *havP, 
  Pointer  imgDataP)
{
  return (
    hav_video_write_ex (havP, imgDataP, NULL, NULL));
}


static void  hav_audio_header_init_ (
  HavAudioHeader  *hdrP,
  Uint32          codecId,
  Uint32          numSamples)
{
  hav_header_init_ ((HavHeader *)hdrP, HAV_AUDIO_HEADER_SIZE, HAV_AUDIO_ID);
  hdrP->dataSize = 0;
  hdrP->codecId = codecId;
  hdrP->numSamples = numSamples;
}


static HavError  hav_audio_write_packet_ (
  Hav      *havP,
  Pointer  headerP,
  Pointer  dataP,
  Uint32   *outputSize)
{
  HavAudioHeader  *hdrP = (HavAudioHeader *) headerP;

  hav_write_header_ (havP, &hdrP->hdr, outputSize);
  *outputSize += f_write (havP->f, &hdrP->dataSize, sizeof(hdrP->dataSize));
  *outputSize += f_write (havP->f, &hdrP->codecId, sizeof(hdrP->codecId));
  *outputSize += f_write (havP->f, &hdrP->numSamples, sizeof(hdrP->numSamples));
  *outputSize += f_write (havP->f, dataP, hdrP->dataSize);
  if (*outputSize != HAV_AUDIO_HEADER_SIZE + hdrP->dataSize)
    return (E_HAV_FWRITE);
  return (E_HAV_OK);
}


HavError  hav_audio_write_ex (
  Hav      *havP, 
  Pointer  sndDataP,
  Uint32   numSamples,
  Uint32   *outputSize)
{
  HavAudioHeader  sndPacket;
  Uint32          oSize;
  int             copyBytes;
  int             bytesLeft = numSamples * havP->aSampleSize;
  int             bytesPerBlock = havP->hdr.aSamplesPerBlock * havP->aSampleSize;
  HavError        result = E_HAV_OK;

  if ((!havP) || (!sndDataP))
    return (E_HAV_NODATA);
  if (outputSize != NULL)
    *outputSize = 0;
  if (havP->aBuf == NULL)
    havP->aBuf = (Pointer) malloc (bytesPerBlock);
  if (sndDataP == NULL) {
    if (havP->aBufUsed > 0) {
      hav_audio_header_init_ (&sndPacket, havP->hdr.aCodecId, havP->aBufUsed / havP->aSampleSize);
      sndPacket.dataSize = havP->aBufUsed;
      sndPacket.hdr.packetSize += sndPacket.dataSize;
      hav_write_adjust_FPT_ (havP, &havP->a, &havP->hdr.aFirstFPT);
      result = hav_audio_write_packet_ (havP, &sndPacket, havP->aBuf, &oSize);
      if (outputSize != NULL)
        *outputSize += oSize;
    }
    return (result);
  }
  havP->hdr.aLen += numSamples;
  while (bytesLeft > 0) {
    copyBytes = bytesPerBlock - havP->aBufUsed;
    if (copyBytes > bytesLeft) {
      memmove ((Uchar *)havP->aBuf + havP->aBufUsed, sndDataP, bytesLeft);
      havP->aBufUsed += bytesLeft;
      break;
    }
    else {
      memmove ((Uchar *)havP->aBuf + havP->aBufUsed, sndDataP, copyBytes);
      sndDataP = (Uchar *)sndDataP + copyBytes;
      bytesLeft -= copyBytes;
      hav_audio_header_init_ (&sndPacket, havP->hdr.aCodecId, havP->hdr.aSamplesPerBlock);
      sndPacket.dataSize = bytesPerBlock;
      sndPacket.hdr.packetSize += sndPacket.dataSize;
      hav_write_adjust_FPT_ (havP, &havP->a, &havP->hdr.aFirstFPT);
      result = hav_audio_write_packet_ (havP, &sndPacket, havP->aBuf, &oSize);
      if (outputSize != NULL)
        *outputSize += oSize;
    }
    havP->aBufUsed = 0;
  }
  havP->a.streamCreated = TRUE;
  return (result);
}


HavError  hav_audio_write (
  Hav      *havP, 
  Pointer  sndDataP, 
  Uint32   numSamples)
{
  return (
    hav_audio_write_ex (havP, sndDataP, numSamples, NULL));
}


HavError  hav_close (
  Hav  *havP)
{
  if (!havP)
    return (E_HAV_NODATA);
  if (havP->v.streamCreated) {
    havP->v.FPT.nextFPT = 0;
    hav_write_FPT_ (havP, &havP->v);
  }
  if (havP->a.streamCreated) {
    hav_audio_write (havP, NULL, 0);
    havP->a.FPT.nextFPT = 0;
    hav_write_FPT_ (havP, &havP->a);
  }
  if (havP->v.streamCreated || havP->a.streamCreated) {
    f_seek (havP->f, 0, SEEK_SET);
    if (f_write (havP->f, &havP->hdr, sizeof(havP->hdr)) != sizeof(havP->hdr))
      return (E_HAV_FWRITE);
    if (havP->palP) {
      free (havP->palP);
      havP->palP = NULL;
    }
  }
  if (havP->vPrevImgDataP)
    free (havP->vPrevImgDataP);
  if (havP->aBuf)
    free (havP->aBuf);
  f_close (havP->f);
  memset (havP, 0, sizeof(Hav));
  return (E_HAV_OK);
}

